function hyp = sjHyperEstimate(X, Y, option)
%
% Hyperparameter estimateϴ Լ 
% 
%
%
%
    if ~isfield(option, 'display')
        option.display = 0;
    end

    option.max_iter  = 1E3;
    option.step_size = 1E-3;
    
    if isfield(option, 'useSE_K')
        hyp = se_estimate(X, Y, option);
    elseif isfield(option, 'useD_K')
        hyp = d_estimate(X, Y, option);
    end

end

%% Squared Exponential Kernel  estimation
function hyp = se_estimate(X, Y, opt)

    % Optimizeϰ ϴ  Ѵ. (ʱⰪ )
    hyp.sigma2f = 1;
    hyp.sigma2x = 1E-2;
    hyp.sigma2w = 1E-3;
    len_X       = size(X, 1);
    
    %  ó ִ´. 
    sigma2f_min = 1E-6;
    sigma2x_min = 1E-6;
    sigma2w_min = 1E-6;
    
    sigma2f_max = 1E1;
    sigma2x_max = 1E-1;
    sigma2w_max = 1E-1;
    
    % Loop 鼭 optimizeѴ. 
    for iter = 1:opt.max_iter
        
        % 0.  update  Ѽ ٽ K Ѵ. 
        opt.sigma2f = hyp.sigma2f;
        opt.sigma2x = hyp.sigma2x;
        opt.sigma2w = hyp.sigma2w;
        K           = sjGP_kernel(X, X, opt) ...
            + opt.sigma2w*eye(len_X, len_X);
        int_K       = inv(K);
        inv_K_Y     = K\Y;
        
        % 1. d(K)/d(sigma2f)
        opt_sig2f   = opt;
        opt_sig2f.sigma2f = 1; % <== K sigma2f ̺  1 ȴ. 
        d_K_d_sigma2f = sjGP_kernel(X, X, opt_sig2f);
        
        % 2. d(K)/d(sigma2w)
        d_K_d_sigma2w = eye(len_X, len_X);
        
        % 3. d(K)/d(sigma2x)
        d_K_d_sigma2x = fun_d_K_d_sigma2f(X, opt);
        
        % 4.  hyperparameter grad Ѵ. 
        grad_sigma2f = 0.5*trace((inv_K_Y*inv_K_Y'-int_K)*d_K_d_sigma2f);
        grad_sigma2w = 0.5*trace((inv_K_Y*inv_K_Y'-int_K)*d_K_d_sigma2w);
        grad_sigma2x = 0.5*trace((inv_K_Y*inv_K_Y'-int_K)*d_K_d_sigma2x);
        
        grad_sigma2w = grad_sigma2w * 1E-4; % <==  scaling 
        
        % 5.  update Ѵ. (ּҰ տ )
        hyp.sigma2f = update_minmax(hyp.sigma2f ...
            , opt.step_size*grad_sigma2f ,sigma2f_min, sigma2f_max);
        hyp.sigma2x = update_minmax(hyp.sigma2x ...
            , opt.step_size*grad_sigma2x ,sigma2x_min, sigma2x_max);
        hyp.sigma2w = update_minmax(hyp.sigma2w ...
            , opt.step_size*grad_sigma2w ,sigma2w_min, sigma2w_max);
        
        % 6.   ȮѴ. 
        
        % 7. Display
        if isfield(opt, 'display')
            if opt.display 
            fprintf(1, '[%d/%d] sigma2f: %.4f(%.6f) sigma2x: %.4f(%.6f) sigma2w: %.4f(%.6f) \n'...
                , iter, opt.max_iter...
                , hyp.sigma2f, grad_sigma2f ...
                , hyp.sigma2x, grad_sigma2x ...
                , hyp.sigma2w, grad_sigma2w );
            end
        end
    end
end

function d_K_d_sigma2f = fun_d_K_d_sigma2f(X, option)
    len_X       = size(X, 1);
    K2          = zeros(len_X, len_X);
    for i = 1:len_X
        for j = 1:len_X 
            x1 = X(i, :);
            x2 = X(j, :);
            K2(i, j) = norm(x1-x2)/2/(option.sigma2x)^2;
        end
    end
    
    d_K_d_sigma2f = sjGP_kernel(X, X, option).*K2;
end

%% Distributed Kernel  estimation 
function hyp = d_estimate(X, Y, opt)
       % Optimizeϰ ϴ  Ѵ. 
    hyp.sigma2f = 1;
    hyp.sigma2w = 1E-3;
    len_X       = size(X, 1);
    
    %  ó ִ´. 
    sigma2f_min = 1E-6;
    sigma2w_min = 1E-6;
    
    sigma2f_max = 1E1;
    sigma2w_max = 1E1;
    
    % Loop 鼭 optimizeѴ. 
    for iter = 1:opt.max_iter
        opt.sigma2f = hyp.sigma2f;
        opt.sigma2w = hyp.sigma2w;
        K           = sjGP_kernel(X, X, opt) ...
            + opt.sigma2w*eye(len_X, len_X);
        int_K       = inv(K);
        inv_K_Y     = K\Y;
        
        % 1. d(K)/d(sigma2f)
        opt_sig2f   = opt;
        opt_sig2f.sigma2f = 1; % <== K sigma2f ̺  1 ȴ. 
        d_K_d_sigma2f = sjGP_kernel(X, X, opt_sig2f);
        
        % 2. d(K)/d(sigma2w)
        d_K_d_sigma2w = eye(len_X, len_X);

        % 3.  hyperparameter grad Ѵ. 
        grad_sigma2f = 0.5*trace((inv_K_Y*inv_K_Y'-int_K)*d_K_d_sigma2f);
        grad_sigma2w = 0.5*trace((inv_K_Y*inv_K_Y'-int_K)*d_K_d_sigma2w);
        
        % 5.  update Ѵ. (ּҰ տ )
        hyp.sigma2f = update_minmax(hyp.sigma2f ...
            , opt.step_size*grad_sigma2f ,sigma2f_min, sigma2f_max);
        hyp.sigma2w = update_minmax(hyp.sigma2w ...
            , opt.step_size*grad_sigma2w ,sigma2w_min, sigma2w_max);
        
        % 6.   ȮѴ. 
        
        % 7. Display
        if isfield(opt, 'display')
            if opt.display
                fprintf(1, '[%d/%d]sigma2f: %.4f sigma2w: %.4f \n'...
                    , iter, opt.max_iter...
                    , hyp.sigma2f, hyp.sigma2w);
            end
        end
    end
end

%% Utillity functions 
function d = btw_min_max(d_org, d_update, d_min, d_max)
    if d_min <= d_update && d_update <= d_max
        d = d_update;
    else
        d = d_org;
    end
end

function d = update_minmax(d_org, d_diff, d_min, d_max)
% update  min~max ̿   updateѴ. 
    d_update = d_org + d_diff;
    if d_min <= d_update && d_update <= d_max
        d = d_update;
    else
        d = d_org;
    end
end
